#!/usr/bin/env python3

"""
Tool for asking a question about the simulation results.
"""

from typing import Type, Optional
from dataclasses import dataclass
import streamlit as st
from pydantic import BaseModel, Field
from langchain_core.tools.base import BaseTool
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain.agents.agent_types import AgentType
from langchain_experimental.agents import create_pandas_dataframe_agent
from langchain_openai import ChatOpenAI
from ..models.basico_model import BasicoModel

@dataclass
class ModelData:
    """
    Dataclass for storing the model data.
    """
    modelid: Optional[int] = None
    sbml_file_path: Optional[str] = None
    model_object: Optional[BasicoModel] = None

class AskQuestionInput(BaseModel):
    """
    Input schema for the AskQuestion tool.
    """
    question: str = Field(description="question about the simulation results")

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class AskQuestionTool(BaseTool):
    """
    Tool for calculating the product of two numbers.
    """
    name: str = "ask_question"
    description: str = "A tool to ask question about the simulation results."
    args_schema: Type[BaseModel] = AskQuestionInput
    return_direct: bool = True
    st_session_key: str = None
    sys_bio_model: ModelData = ModelData()

    def _run(self,
             question: str,
             run_manager: Optional[CallbackManagerForToolRun] = None) -> str:
        """
        Run the tool.

        Args:
            question (str): The question to ask about the simulation results.
            run_manager (Optional[CallbackManagerForToolRun]): The CallbackManagerForToolRun object.

        Returns:
            str: The answer to the question.
        """
        st_session_key = self.st_session_key
        sys_bio_model = self.sys_bio_model
        # Check if sys_bio_model is provided in the input
        if sys_bio_model.modelid or sys_bio_model.sbml_file_path or sys_bio_model.model_object:
            if sys_bio_model.modelid is not None:
                model_object = BasicoModel(model_id=sys_bio_model.modelid)
            elif sys_bio_model.sbml_file_path is not None:
                model_object = BasicoModel(sbml_file_path=sys_bio_model.sbml_file_path)
            else:
                model_object = sys_bio_model.model_object
        else:
            # If the sys_bio_model is not provided in the input,
            # get it from the Streamlit session state
            if st_session_key:
                if st_session_key not in st.session_state:
                    return f"Session key {st_session_key} not found in Streamlit session state."
                model_object = st.session_state[st_session_key]
            else:
                return "Please provide a valid model object or \
                    Streamlit session key that contains the model object."
        # Update the object in the streamlit session state
        if st_session_key:
            st.session_state[st_session_key] = model_object
        if model_object.simulation_results is None:
            model_object.simulate()
        df = model_object.simulation_results
        # If there is a Streamlit session key,
        # display the simulation results
        if st_session_key:
            st.text(f"Simulation Results of the model {model_object.model_id}")
            st.dataframe(df, use_container_width = False, width = 650)
        # Check if run_manager's metadata has the key 'prompt_content'
        prompt_content = None
        if run_manager and 'prompt' in run_manager.metadata:
            prompt_content = run_manager.metadata['prompt']
        # Create a pandas dataframe agent with OpenAI
        df_agent = create_pandas_dataframe_agent(ChatOpenAI(model="gpt-3.5-turbo"),
                                                  allow_dangerous_code=True,
                                                  agent_type=AgentType.OPENAI_FUNCTIONS,
                                                  df=df,
                                                  prefix=prompt_content)
        llm_result = df_agent.invoke(question)
        return llm_result["output"]

    def get_metadata(self):
        """
        Get metadata for the tool.

        Returns:
            dict: The metadata for the tool.
        """
        return {
            "name": self.name,
            "description": self.description
        }



'''
This file is used to import all the modules in the package.
'''
# import everything from the module
from . import ask_question
from . import simulate_model
from . import custom_plotter
from . import fetch_parameters
from . import model_description
from . import search_models



#!/usr/bin/env python3

"""
Tool for searching models based on search query.
"""

from urllib.error import URLError
from time import sleep
from typing import Type
from pydantic import BaseModel, Field
from basico import biomodels
from langchain_core.tools import BaseTool
from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI

class SearchModelsInput(BaseModel):
    """
    Input schema for the search models tool.
    """
    query: str = Field(description="Search models query", default=None)

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class SearchModelsTool(BaseTool):
    """
    Tool for returning the search results based on the search query.
    """
    name: str = "search_models"
    description: str = "Search models based on search query."
    args_schema: Type[BaseModel] = SearchModelsInput
    return_direct: bool = True

    def _run(self, query: str) -> str:
        """
        Run the tool.

        Args:
            query (str): The search query.

        Returns:
            str: The answer to the question.
        """
        attempts = 0
        max_retries = 3
        while attempts < max_retries:
            try:
                search_results = biomodels.search_for_model(query)
                break
            except URLError as e:
                attempts += 1
                sleep(10)
                if attempts >= max_retries:
                    raise e
        llm = ChatOpenAI(model="gpt-4o-mini")
        # Check if run_manager's metadata has the key 'prompt_content'
        prompt_content = f'''
                        Convert the input into a table.

                        The table must contain the following columns:
                        - #
                        - BioModel ID
                        - BioModel Name
                        - Format
                        - Submission Date

                        Additional Guidelines:
                        - The column # must contain the row number starting from 1.
                        - Embed the url for each BioModel ID in the table 
                        in the first column in the markdown format.
                        - The Submission Date must be in the format YYYY-MM-DD.

                        Input:
                        {input}.
                        '''
        prompt_template = ChatPromptTemplate.from_messages(
            [("system", prompt_content),
             ("user", "{input}")]
        )
        parser = StrOutputParser()
        chain = prompt_template | llm | parser
        return chain.invoke({"input": search_results})

    def get_metadata(self):
        """
        Get metadata for the tool.

        Returns:
            dict: The metadata for the tool.
        """
        return {
            "name": self.name,
            "description": self.description
        }



#!/usr/bin/env python3

"""
Tool for asking a question about the model description.
"""

from typing import Type, Optional
from dataclasses import dataclass
from pydantic import BaseModel, Field, model_validator
import streamlit as st
from langchain_core.tools import BaseTool
from langchain_core.output_parsers import StrOutputParser
from langchain_core.callbacks import CallbackManagerForToolRun
from langchain_core.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from ..models.basico_model import BasicoModel

@dataclass
class ModelData:
    """
    Dataclass for storing the model data.
    """
    model_id: Optional[int] = None
    sbml_file_path: Optional[str] = None
    model_object: Optional[BasicoModel] = None

    # Check if model_object is an instance of BasicoModel
    # otherwise make it None. This is important because
    # sometimes the LLM may invoke the tool with an
    # inappropriate model_object.
    @model_validator(mode="before")
    @classmethod
    def check_model_object(cls, data):
        """
        Check if model_object is an instance of BasicoModel.
        """
        if 'model_object' in data:
            if not isinstance(data['model_object'], BasicoModel):
                data['model_object'] = None
        return data

class ModelDescriptionInput(BaseModel):
    """
    Input schema for the ModelDescription tool.
    """
    question: str = Field(description="question about the model description")
    sys_bio_model: ModelData = Field(description="model data", default=None)

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class ModelDescriptionTool(BaseTool):
    """
    Tool for returning the description of the specified BioModel.
    """
    name: str = "model_description"
    description: str = '''A tool to ask about the description of the model.'''
    args_schema: Type[BaseModel] = ModelDescriptionInput
    return_direct: bool = True
    st_session_key: str = None

    def _run(self,
             question: str,
             sys_bio_model: ModelData = ModelData(),
             run_manager: Optional[CallbackManagerForToolRun] = None) -> str:
        """
        Run the tool.

        Args:
            question (str): The question to ask about the model description.
            run_manager (Optional[CallbackManagerForToolRun]): The CallbackManagerForToolRun object.

        Returns:
            str: The answer to the question.
        """
        st_session_key = self.st_session_key
        # Check if sys_bio_model is provided in the input schema
        if sys_bio_model.model_id or sys_bio_model.sbml_file_path \
            or sys_bio_model.model_object not in [None, "", {}]:
            if sys_bio_model.model_id:
                model_object = BasicoModel(model_id=sys_bio_model.model_id)
            elif sys_bio_model.sbml_file_path:
                model_object = BasicoModel(sbml_file_path=sys_bio_model.sbml_file_path)
            else:
                print (sys_bio_model.model_object, 'model_object')
                model_object = sys_bio_model.model_object
            if st_session_key:
                st.session_state[st_session_key] = model_object
        # Check if sys_bio_model is provided in the Streamlit session state
        elif st_session_key:
            if st_session_key not in st.session_state:
                return f"Session key {st_session_key} " \
                        "not found in Streamlit session state."
            model_object = st.session_state[st_session_key]
        else:
            return "Please provide a valid model object or Streamlit "\
                "session key that contains the model object."
        # check if model_object is None
        if model_object is None:
            return "Please provide a BioModels ID or an SBML file path for the model."
        description = model_object.description
        if description in [None, ""]:
            return "No description found for the model."
        # Append the BioModel ID of the model to the description
        description = f"{description} (BioModel ID: {model_object.model_id})"
        llm = ChatOpenAI(model="gpt-3.5-turbo")
        # Check if run_manager's metadata has the key 'prompt_content'
        if run_manager and 'prompt' in run_manager.metadata:
            prompt_content = run_manager.metadata['prompt']
        else:
            prompt_content = '''
                            Given the description of a System biology model:
                            {description},
                            answer the user question:
                            {question}.
                            '''
        prompt_template = ChatPromptTemplate.from_messages(
            [("system", prompt_content),
             ("user", "{description} {question}")]
        )
        parser = StrOutputParser()
        chain = prompt_template | llm | parser
        return chain.invoke({"description": description,
                             "question": question})

    def get_metadata(self):
        """
        Get metadata for the tool.

        Returns:
            dict: The metadata for the tool.
        """
        return {
            "name": self.name,
            "description": self.description
        }



#!/usr/bin/env python3

"""
Tool for simulating a model.
"""

from typing import Type, Union, List, Optional
from dataclasses import dataclass
import basico
from pydantic import BaseModel, Field
from langchain_core.tools import BaseTool
import streamlit as st
# import plotly.express as px
from ..models.basico_model import BasicoModel

@dataclass
class ModelData:
    """
    Dataclass for storing the model data.
    """
    modelid: Optional[int] = None
    sbml_file_path: Optional[str] = None

@dataclass
class TimeData:
    """
    Dataclass for storing the time data.
    """
    duration: Union[int, float] = 100
    interval: Union[int, float] = 10

@dataclass
class SpeciesData:
    """
    Dataclass for storing the species data.
    """
    species_name: List[str] = None
    species_concentration: List[Union[int, float]] = None

@dataclass
class TimeSpeciesNameConcentration:
    """
    Dataclass for storing the time, species name, and concentration data.
    """
    time: Union[int, float] = None
    species_name: str = None
    species_concentration: Union[int, float] = None

@dataclass
class RecurringData:
    """
    Dataclass for storing the species and time data 
    on recurring basis.
    """
    data: List[TimeSpeciesNameConcentration] = None

class SimulateModelInput(BaseModel):
    """
    Input schema for the SimulateModel tool.
    """
    model_data: ModelData = Field(description="model data", default=None)
    time_data: TimeData = Field(description="time data", default=None)
    species_data: SpeciesData = Field(description="species data", default=None)
    recurring_data: RecurringData = Field(description="recurring data", default=None)

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class SimulateModelTool(BaseTool):
    """
    Tool for simulating a model.
    """
    name: str = "simulate_model"
    description: str = "A tool to simulate a model."
    args_schema: Type[BaseModel] = SimulateModelInput
    st_session_key: str = None

    def _run(self,
                model_data: ModelData = None,
                time_data: TimeData = None,
                species_data: SpeciesData = None,
                recurring_data: RecurringData = None):
        """
        Run the tool.

        Args:
            model_data (Optional[ModelData]): The model data.
            time_data (Optional[TimeData]): The time data.
            species_data (Optional[SpeciesData]): The species data.
            recurring_data (Optional[RecurringData]): The recurring data.

        Returns:
            str: The result of the simulation.
        """
        # st_session_key = self.st_session_key
        # Retrieve the model ID, duration, and interval
        modelid = model_data.modelid if model_data is not None else None
        # duration = time_data.duration if time_data is not None else 100.0
        # interval = time_data.interval if time_data is not None else 10
        # Prepare the dictionary of species data
        # that will be passed to the simulate method
        # of the BasicoModel class
        # dic_species_data = None
        # if species_data is not None:
        #     dic_species_data = dict(zip(species_data.species_name,
        #                             species_data.species_concentration))
        dic_species_data = dict(zip(species_data.species_name, \
            species_data.species_concentration)) \
                if species_data is not None else None
        # Retrieve the SBML file path from the Streamlit session state
        # otherwise retrieve it from the model_data object if the user
        # has provided it.
        sbml_file_path = model_data.sbml_file_path if model_data is not None else None
        if self.st_session_key:
            if self.st_session_key not in st.session_state:
                return f"Session key {self.st_session_key} not found in Streamlit session state."
            if 'sbml_file_path' in st.session_state:
                sbml_file_path = st.session_state.sbml_file_path
        # Check if both modelid and sbml_file_path are None
        if modelid is None and sbml_file_path is None:
            # Then load the model from the Streamlit session state
            # if the streamlit session exists
            if self.st_session_key:
                model_object = st.session_state[self.st_session_key]
                # If this model object is None, then return an error message
                if model_object is None:
                    return "Please provide a BioModels ID or an SBML file path for simulation."
                # Retrieve the model ID from the model object
                modelid = model_object.model_id
            else:
                # Otherwise return an error message
                return "Please provide a BioModels ID or an SBML file path for simulation."
        elif modelid:
            # Create a BasicoModel object with the model ID
            # model_object = BasicoModel(model_id=modelid)
            model_object = BasicoModel(model_id=modelid)
            # Save the model object in the Streamlit session state
            st.session_state[self.st_session_key] = model_object
        elif sbml_file_path:
            # Create a BasicoModel object with the SBML file path
            model_object = BasicoModel(sbml_file_path=sbml_file_path)
            modelid = model_object.model_id
            # Save the model object in the Streamlit session state
            st.session_state[self.st_session_key] = model_object
        # Add recurring events (if any) to the model
        if recurring_data is not None:
            for row in recurring_data.data:
                tp, sn, sc = row.time, row.species_name, row.species_concentration
                basico.add_event(f'{sn}_{tp}',
                                 f'Time > {tp}',
                                 [[sn, str(sc)]],
                                 model=model_object.copasi_model)
                # print (f'Added event {sn}_{tp} at time {tp} \
                #        for species {sn} with concentration {sc}')
        # Simulate the model
        df = model_object.simulate(
            parameters=dic_species_data,
            duration=time_data.duration if time_data is not None else 100.0,
            interval=time_data.interval if time_data is not None else 10)
        # Convert the DataFrame to long format for plotting
        # and ignore the index column
        df = df.melt(id_vars='Time',
                    var_name='Species',
                    value_name='Concentration')
        # Plot the simulation results using Plotly
        # fig = px.line(df,
        #                 x='Time',
        #                 y='Concentration',
        #                 color='Species',
        #                 title=f"Concentration of Species over Time in the model {modelid}",
        #                 height=600,
        #                 width=800
        # )
        # Display the plot in Streamlit
        # st.plotly_chart(fig, use_container_width = False)
        # if modelid is None:
        #     modelid = "internal"
        # return f"Simulation results for the model {modelid}."
        return "Simulation results for the model."

    def get_metadata(self):
        """
        Get metadata for the tool.

        Returns:
            dict: The metadata for the tool.
        """
        return {
            "name": self.name,
            "description": self.description
        }
